#!/usr/bin/env python
# Copyright (C) 2018
# Associated Universities, Inc. Washington DC, USA.
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be addressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA

"""CASAmpi Python Module

This is a standard python module that provides the CASA MPI framework to process tasks
and tools in parallel.
"""
from __future__ import division, print_function

import zlib
import sys
import os

try:
    import casatools
    from casatools.config import build as tools_config
except ImportError as exc:
    print("cannot find CASAtools "
          "(https://open-bitbucket.nrao.edu/projects/CASA/repos/casatools/browse) in "
          "PYTHONPATH. Exception found: {}".format(exc))
    os._exit(1)

from setuptools import setup, Extension, find_packages, Command

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'src'))

classifiers = """\
Development Status :: 3 - Alpha
Intended Audience :: Developers
License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)
Programming Language :: Python :: 2.7
Programming Language :: C++
Topic :: Software Development
Topic :: Scientific/Engineering :: Astronomy
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: MacOS :: MacOS X
Operating System :: POSIX
"""

keywords = """
CASA
common astronomy software applications
scientific computing
MPI
message passing interface
parallel computing
"""

module_name = 'casampi'
pyversion = float(sys.version_info[0]) + float(sys.version_info[1]) / 10.0

# hardcoded for the moment as the one in casatasks use casa-source
casampi_version = '0.5.6'

if pyversion < 3:
    str_encode = str
    str_decode = str
    def pipe_decode(output):
        return output
else:
    def str_encode(s):
        return bytes(s,sys.getdefaultencoding())
    def str_decode(bs):
        return bs.decode(sys.getdefaultencoding(),"strict")
    def pipe_decode(output):
        if isinstance(output,bytes) or isinstance(output,bytearray):
            return str_decode(output)
        elif isinstance(output,tuple):
            return (str_decode(output[0]),str_decode(output[1]))
        else:
            return ("","")

# Build wheel
cmd_setup = {}
try:
    from wheel.bdist_wheel import bdist_wheel
    cmd_setup['bdist_wheel'] = bdist_wheel
except ImportError:
    pass  # custom command not needed if wheel is not installed


setup(name=module_name,
      version=casampi_version,
      maintainer="Sandra Castro",
      maintainer_email="scastro@eso.org",
      author="CASA development team",
      author_email="casa-feedback@nrao.edu",
      url="https://open-bitbucket.nrao.edu/projects/CASA/repos/casampi/browse",
      download_url="https://casa.nrao.edu/download/",
      license="GNU Library or Lesser General Public License (LGPL)",
      classifiers=[ 'Programming Language :: Python :: 3',
                    'Programming Language :: Python :: 2.7'],
      keywords=keywords,
      description="the CASA MPI framework",
      long_description="The CASAmpi is the MPI framework to process CASA tools and tasks in parallel.",
      test_suite='casampi.tests',
      cmdclass=cmd_setup,
      packages=find_packages(where='src'),
      package_dir = {'': 'src'},
      install_requires=["mpi4py >= 2.0", "casatools"]
)
